#!/bin/bash

# (The MIT License)
#
# Copyright (c) 2018 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


_blackhole_complete() {
    local cur_word=${COMP_WORDS[COMP_CWORD]}
    local options="-b -c -d -h -ls -q -t -v
                   --background --conf --debug --help --less-secure --quiet
                   --test --version"
    if [[ "$cur_word" == -*  ]]; then
        # shellcheck disable=SC2207
        COMPREPLY=( $( compgen -W "$options" -- "$cur_word" ) )
    fi
    return 0
}

complete -o default -F _blackhole_complete blackhole
